# Copyright 2026 Xiaohongshu. (Author: Kaituo Xu, Kai Huang)

SAMPLE_RATE = 16000
FRAME_LENGTH_MS = 25
FRAME_SHIFT_MS = 10
FRAME_LENGTH_S = 0.025
FRAME_SHIFT_S = 0.010
FRAME_LENGTH_SAMPLE = int(SAMPLE_RATE * FRAME_LENGTH_MS / 1000)
FRAME_SHIFT_SAMPLE = int(SAMPLE_RATE * FRAME_SHIFT_MS / 1000)
FRAME_PER_SECONDS = int(1000 / FRAME_SHIFT_MS)
